import static org.junit.Assert.*;

import org.junit.Test;

public class PublicTests {

    // Allowed delta for comparisons
    private static final double EPSILON = 0.001;

    // ################
    // SIMILARITY TESTS
    // ################

    @Test(expected = IllegalArgumentException.class)
    public void testSimilarity_firstVectorNull_shouldThrowException() {
        byte[] ratings1 = null;
        byte[] ratings2 = { 1 };

        FilmRecommender.similarity(ratings1, ratings2);
    }

    @Test(expected = IllegalArgumentException.class)
    public void testSimilarity_secondVectorNull_shouldThrowException() {
        byte[] ratings1 = { 1 };
        byte[] ratings2 = null;

        FilmRecommender.similarity(ratings1, ratings2);
    }

    @Test(expected = IllegalArgumentException.class)
    public void testSimilarity_vectorsUnequalLength_shouldThrowException() {
        byte[] ratings1 = { 1, 2 };
        byte[] ratings2 = { 1 };

        FilmRecommender.similarity(ratings1, ratings2);
    }

    @Test
    public void testSimilarity_emptyRatings_shouldReturn0() {
        byte[] ratings1 = {};
        byte[] ratings2 = {};

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_singleRating_shouldReturn0() {
        byte[] ratings1 = { 1 };
        byte[] ratings2 = { 2 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_noCommonRatings_shouldReturn0() {
        byte[] ratings1 = { 1, -3, 1 };
        byte[] ratings2 = { -3, 2, -3 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_oneCommonRatings_shouldReturn0() {
        byte[] ratings1 = { 1, -3, 1 };
        byte[] ratings2 = { 2, 2, -3 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_zeroRatings_shouldReturn0() {
        byte[] ratings1 = { 0, 0, 0 };
        byte[] ratings2 = { 2, 2, -3 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_similarRatings_shouldReturnHighValue() {
        byte[] ratings1 = { 1, 2, 1 };
        byte[] ratings2 = { 1, 2, 1 };

        assertEquals(1.0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_similarRatings2_shouldReturnHighValue() {
        byte[] ratings1 = { 1, 2, 2 };
        byte[] ratings2 = { 1, 2, 1 };

        assertEquals(7 / (3 * Math.sqrt(6)), FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_similarRatings3_shouldReturnHighValue() {
        byte[] ratings1 = { 1, -3, 1 };
        byte[] ratings2 = { 2, 2, 2 };

        assertEquals(1.0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_orthogonalRatings_shouldReturn0() {
        byte[] ratings1 = { 1, 0, 0 };
        byte[] ratings2 = { 0, 1, 0 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_orthogonalRatings2_shouldReturn0() {
        byte[] ratings1 = { 1, 0, 1 };
        byte[] ratings2 = { 0, 1, 0 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_orthogonalRatings3_shouldReturn0() {
        byte[] ratings1 = { 1, 1, 1 };
        byte[] ratings2 = { -1, -1, 2 };

        assertEquals(0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_dissimilarRatings_shouldReturnLowValue() {
        byte[] ratings1 = { 1, 1, 1 };
        byte[] ratings2 = { -2, -2, -2 };

        assertEquals(-1.0, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_dissimilarRatings2_shouldReturnLowValue() {
        byte[] ratings1 = { -1, -2, 1 };
        byte[] ratings2 = { 2, 2, 1 };

        assertEquals(-5.0 / (3 * Math.sqrt(6)), FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }

    @Test
    public void testSimilarity_dissimilarRatings3_shouldReturnLowValue() {
        byte[] ratings1 = { -1, 2, -1 };
        byte[] ratings2 = { 1, 2, 1 };

        assertEquals(2.0 / 6, FilmRecommender.similarity(ratings1, ratings2), EPSILON);
    }


    // ####################
    // RECOMMENDATION TESTS
    // ####################


    @Test
    public void testComputeRatings_noOtherUsers_shouldReturnOriginalVector() {
        byte[][] ratings = {{-3, -2, -1, 0, 1, 2}};
        double[] expectedRatings = {-3, -2, -1, 0, 1, 2};

        double[] computedRatings = FilmRecommender.computeRatings(ratings, 0);

        assertArrayEquals(expectedRatings, computedRatings, EPSILON);
    }

    @Test
    public void testComputeRecommendation_noOtherUsers_shouldFirstUnseenIndex() {
        byte[][] ratings = {{-3, -2, -1, 0, 1, 2}};

        int recommendation = FilmRecommender.computeRecommendation(ratings, 0);

        assertEquals(0, recommendation, EPSILON);
    }

    // Similarities zero
    @Test
    public void testComputeRatings_similaritiesZero_shouldReturnAverageRating() {
        byte[][] ratings = {{-3, -3, 1, 0, 1, 1}, {-2, 2, -1, 0, 1, 0}, {-1, 1, 1, 0, 1, -2}};
        double[] expectedRatings = {-1.5, 1.5, 0, 0, 0, 0};

        double[] computedRatings = FilmRecommender.computeRatings(ratings, 0);

        assertArrayEquals(expectedRatings, computedRatings, EPSILON);
    }

    @Test
    public void testComputeRecommendation_similaritiesZero_shouldChooseAccordingsToAverageRating() {
        byte[][] ratings = {{-3, -3, 1, 0, 1, 1}, {-2, 2, -1, 0, 1, 0}, {-1, 1, 1, 0, 1, -2}};

        int recommendation = FilmRecommender.computeRecommendation(ratings, 0);

        assertEquals(1, recommendation, EPSILON);
    }

    // No one has seen a particular film
    @Test
    public void testComputeRatings_noOneHasSeenFilm_shouldReturn0ForThatFilm() {
        byte[][] ratings = {{-3, -3, 1, 0, 1, 1}, {-3, -2, 1, 0, 1, 0}, {-3, 1, 1, 0, 1, 0}};
        double[] expectedRatings = {0, -0.5, 0, 0, 0, 0};

        double[] computedRatings = FilmRecommender.computeRatings(ratings, 0);

        assertArrayEquals(expectedRatings, computedRatings, EPSILON);
    }

    @Test
    public void testComputeRecommendation_noOneHasSeenFilm_shouldChooseAccordingsToAverageRating() {
        byte[][] ratings = {{-3, -3, 1, 0, 1, 1}, {-3, -2, 1, 0, 1, 0}, {-3, 1, 1, 0, 1, 0}};

        int recommendation = FilmRecommender.computeRecommendation(ratings, 0);

        assertEquals(0, recommendation, EPSILON);
    }

    // The user has seen all films
    @Test
    public void testComputeRatings_userHasSeenAllFilms_shouldReturnAllRatings0() {
        byte[][] ratings = {{1, 1, 1, 0, 1, 1}, {-3, -2, 1, 0, 1, 0}, {-3, 1, 1, 0, 1, 0}};
        double[] expectedRatings = {0, 0, 0, 0, 0, 0};

        double[] computedRatings = FilmRecommender.computeRatings(ratings, 0);

        assertArrayEquals(expectedRatings, computedRatings, EPSILON);
    }

    @Test
    public void testComputeRecommendation_userHasSeenAllFilms_shouldReturnMinus1() {
        byte[][] ratings = {{1, 1, 1, 0, 1, 1}, {-2, 2, -1, 0, 1, 0}, {-1, 1, 1, 0, 1, -2}};

        int recommendation = FilmRecommender.computeRecommendation(ratings, 0);

        assertEquals(-1, recommendation, EPSILON);
    }

    // Small examples
    @Test
    public void testComputeRatings_smallSample1_shouldReturnNormalRatings() {
        byte[][] ratings = {{-3, 2, -3, -1}, {-1, -3, 2, 1}, {2, 1, -2, -2}, {1, 2, -3, -1}, {-2, -1, 1, 1},
                {2, -3, -1, -3}};

        // Test user 0
        double[] expectedRatings0 = {(2.6 + 6 / Math.sqrt(10)) / (1.8 + 3 / Math.sqrt(10)), 0,
                (-1.6 - 3 / Math.sqrt(10)) / (0.8 + 3 / Math.sqrt(10)), 0};
        double[] computedRatings0 = FilmRecommender.computeRatings(ratings, 0);
        assertArrayEquals(expectedRatings0, computedRatings0, EPSILON);

        // Test user 1
        double[] expectedRatings1 = {0, -(4 * Math.sqrt(2) + 17) / (4 * Math.sqrt(2) + 11), 0, 0};
        double[] computedRatings1 = FilmRecommender.computeRatings(ratings, 1);
        assertArrayEquals(expectedRatings1, computedRatings1, EPSILON);

        // Test user 2
        double[] expectedRatings2 = {0, 0, 0, 0};
        double[] computedRatings2 = FilmRecommender.computeRatings(ratings, 2);
        assertArrayEquals(expectedRatings2, computedRatings2, EPSILON);

        // Test user 3
        double[] expectedRatings3 = {0, 0,
                -(2 + 2 * Math.sqrt(2.0 / 3) + 5.0 / 6) / (1 + Math.sqrt(2.0 / 3) + 5.0 / 6), 0};
        double[] computedRatings3 = FilmRecommender.computeRatings(ratings, 3);
        assertArrayEquals(expectedRatings3, computedRatings3, EPSILON);

        // Test user 4
        double[] expectedRatings4 = {0, 0, 0, 0};
        double[] computedRatings4 = FilmRecommender.computeRatings(ratings, 4);
        assertArrayEquals(expectedRatings4, computedRatings4, EPSILON);

        // Test user 5
        double[] expectedRatings5 = {0, 1, 0, -(6 / Math.sqrt(10) + 1.8) / (3 / Math.sqrt(10) + 1.8)};
        double[] computedRatings5 = FilmRecommender.computeRatings(ratings, 5);
        assertArrayEquals(expectedRatings5, computedRatings5, EPSILON);
    }

    @Test
    public void testComputeRecommendation_smallSample1_shouldReturnNormalRecommendation() {
        byte[][] ratings = {{-3, 2, -3, -1}, {-1, -3, 2, 1}, {2, 1, -2, -2}, {1, 2, -3, -1}, {-2, -1, 1, 1},
                {2, -3, -1, -3}};

        // Test user 0
        assertEquals(0, FilmRecommender.computeRecommendation(ratings, 0), EPSILON);

        // Test user 1
        assertEquals(1, FilmRecommender.computeRecommendation(ratings, 1), EPSILON);

        // Test user 2
        assertEquals(-1, FilmRecommender.computeRecommendation(ratings, 2), EPSILON);

        // Test user 3
        assertEquals(2, FilmRecommender.computeRecommendation(ratings, 3), EPSILON);

        // Test user 4
        assertEquals(-1, FilmRecommender.computeRecommendation(ratings, 4), EPSILON);

        // Test user 5
        assertEquals(1, FilmRecommender.computeRecommendation(ratings, 5), EPSILON);
    }

    @Test
    public void testComputeRatings_smallSample2_shouldReturnNormalRatings() {
        byte[][] ratings = {{-3, -3, -1, 2}, {-3, 1, -3, -1}, {-3, -3, -3, -3}, {1, 2, -2, -1}};

        // Test user 0
        double[] expectedRatings0 = {1, 1.5, 0, 0};
        double[] computedRatings0 = FilmRecommender.computeRatings(ratings, 0);
        assertArrayEquals(expectedRatings0, computedRatings0, EPSILON);

        // Test user 1
        double[] expectedRatings1 = {1, 0, -2, 0};
        double[] computedRatings1 = FilmRecommender.computeRatings(ratings, 1);
        assertArrayEquals(expectedRatings1, computedRatings1, EPSILON);

        // Test user 2
        double[] expectedRatings2 = {1, 1.5, -1.5, 0};
        double[] computedRatings2 = FilmRecommender.computeRatings(ratings, 2);
        assertArrayEquals(expectedRatings2, computedRatings2, EPSILON);

        // Test user 3
        double[] expectedRatings3 = {0, 0, 0, 0};
        double[] computedRatings3 = FilmRecommender.computeRatings(ratings, 3);
        assertArrayEquals(expectedRatings3, computedRatings3, EPSILON);
    }

    @Test
    public void testComputeRecommendation_smallSample2_shouldReturnNormalRecommendation() {
        byte[][] ratings = {{-3, -3, -1, 2}, {-3, 1, -3, -1}, {-3, -3, -3, -3}, {1, 2, -2, -1}};

        // Test user 0
        assertEquals(1, FilmRecommender.computeRecommendation(ratings, 0), EPSILON);

        // Test user 1
        assertEquals(0, FilmRecommender.computeRecommendation(ratings, 1), EPSILON);

        // Test user 2
        assertEquals(1, FilmRecommender.computeRecommendation(ratings, 2), EPSILON);

        // Test user 3
        assertEquals(-1, FilmRecommender.computeRecommendation(ratings, 3), EPSILON);
    }

    @Test
    public void testComputeRatings_smallSample3_shouldReturnNormalRatings() {
        byte[][] ratings = {{-2, -3, 2, -1, -3}, {1, -3, -2, -3, -1}, {-3, -1, 0, 1, -1}, {1, 2, 2, -3, -1}};

        // Test user 0
        double[] expectedRatings0 = {0,
                (1 / Math.sqrt(5) + 2 / Math.sqrt(10)) / (1 / Math.sqrt(5) + 1 / Math.sqrt(10)), 0, 0,
                (1 / Math.sqrt(5) + 2 / Math.sqrt(10)) / (1 / Math.sqrt(5) + 4 / Math.sqrt(10))};
        double[] computedRatings0 = FilmRecommender.computeRatings(ratings, 0);
        assertArrayEquals(expectedRatings0, computedRatings0, EPSILON);

        // Test user 1
        double[] expectedRatings1 = {0, -(1 / Math.sqrt(5) + 2.0 / 3) / (1 / Math.sqrt(5) + 1.0 / 3), 0, 1, 0};
        double[] computedRatings1 = FilmRecommender.computeRatings(ratings, 1);
        assertArrayEquals(expectedRatings1, computedRatings1, EPSILON);

        // Test user 2
        double[] expectedRatings2 = {
                (3 / Math.sqrt(5) - 1 / (3 * Math.sqrt(2))) / (2 / Math.sqrt(5) + 1 / (3 * Math.sqrt(2))), 0, 0, 0, 0};
        double[] computedRatings2 = FilmRecommender.computeRatings(ratings, 2);
        assertArrayEquals(expectedRatings2, computedRatings2, EPSILON);

        // Test user 3
        double[] expectedRatings3 = {0, 0, 0, -1, 0};
        double[] computedRatings3 = FilmRecommender.computeRatings(ratings, 3);
        assertArrayEquals(expectedRatings3, computedRatings3, EPSILON);
    }

    @Test
    public void testComputeRecommendation_smallSample3_shouldReturnNormalRecommendation() {
        byte[][] ratings = {{-2, -3, 2, -1, -3}, {1, -3, -2, -3, -1}, {-3, -1, 0, 1, -1}, {1, 2, 2, -3, -1}};

        // Test user 0
        assertEquals(1, FilmRecommender.computeRecommendation(ratings, 0), EPSILON);

        // Test user 1
        assertEquals(3, FilmRecommender.computeRecommendation(ratings, 1), EPSILON);

        // Test user 2
        assertEquals(0, FilmRecommender.computeRecommendation(ratings, 2), EPSILON);

        // Test user 3
        assertEquals(3, FilmRecommender.computeRecommendation(ratings, 3), EPSILON);
    }

}